#!/usr/bin/env python
# -*- coding: utf-8 -*-

import functools
import os
import re
import shutil
import time
from contextlib import contextmanager
from pathlib import Path
from typing import List, Set, Union

from clickgen.db import DATA, CursorDB


@contextmanager
def chdir(directory: Union[str, Path]):
    """Temporary change working directory using `with` syntax.

    :param directory: path to directory.
    :type directory: Union[str, pathlib.Path]

    """

    prev_cwd = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def remove_util(p: Union[str, Path]) -> None:
    """Remove this file, directory or symlink.

    :param p: path to directory.
    :type p: Union[str, pathlib.Path]
    :return: None
    :rtype: None

    """
    p_obj: Path = Path(p)

    if p_obj.exists():
        if p_obj.is_dir():
            shutil.rmtree(p_obj)
        else:
            p_obj.unlink()
    else:
        pass


class PNGProvider:
    """Provide organized `.png` files."""

    bitmaps_dir: Path
    __pngs: List[str] = []

    def __init__(self, bitmaps_dir: Union[str, Path]) -> None:
        """Init `PNGProvider`.

        :param bitmaps_dir: Path to directory where `.png` files are stored.
        :type bitmaps_dir: Union[str, Path]

        """
        super().__init__()
        self.bitmaps_dir = Path(bitmaps_dir)
        for f in sorted(self.bitmaps_dir.iterdir()):
            self.__pngs.append(f.name)

        if len(self.__pngs) == 0:
            raise FileNotFoundError(
                f"'*.png' files not found in '{self.bitmaps_dir.absolute()}'"
            )

    def get(self, key: str) -> Union[List[Path], Path]:
        """Retrieve `pathlib.Path` of filtered `.png` file/s.

        This method return file location in `pathlib.Path` object.

        Runtime directory `sync` is **not supported**, Which means creating
        or deleting a file on programs execution is not update class `__pngs`
        state.

        :param key: `key` is filename
        :type key: str
        :return: Returns `pathlib.Path` object or `list` of `pathlib.Path`
                 object/s.
        """

        k = key.split(".")
        if len(k) == 1:
            r = re.compile(fr"^{k[0]}(?:-\d+)?.png$")
        else:
            r = re.compile(fr"^{k[0]}(?:-\d+)?.{k[1]}$")

        matched_pngs = filter(r.match, self.__pngs)

        paths = list(set(map(lambda x: self.bitmaps_dir / x, matched_pngs)))
        if len(paths) == 1:
            return paths[0]
        return paths


def add_missing_xcursors(
    directory: Path,
    data: List[Set[str]] = DATA,
    rename: bool = False,
    force: bool = False,
) -> None:
    """Create symlinks of missing ``Xcursor``"""
    if not directory.exists() or not directory.is_dir():
        raise NotADirectoryError(directory.absolute())

    db: CursorDB = CursorDB(data)

    # Removing all symlinks cursors
    if force:
        for xcursor in directory.iterdir():
            if xcursor.is_symlink():
                xcursor.unlink(missing_ok=True)

    xcursors = sorted(directory.iterdir())

    for xcursor in xcursors:
        # Rename Xcursor according to Database, If necessary
        if rename:
            new_path = db.rename_file(xcursor)
            if new_path:
                xcursor = xcursor.rename(new_path)

        # Creating symlinks
        links = db.search_symlinks(xcursor.stem)
        if links:
            for link in links:
                with chdir(directory):
                    os.symlink(xcursor, link)


#
# Useful for development
#
def timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()

        value = func(*args, **kwargs)

        end_time = time.perf_counter()
        run_time = end_time - start_time
        print(f"Finished {func.__name__!r} in {run_time:.6f} secs")

        return value

    return wrapper_timer


def debug(func):
    """Print the function signature and return value"""

    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [repr(a) for a in args]  # 1
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]  # 2
        signature = ", ".join(args_repr + kwargs_repr)  # 3

        print(f"Calling {func.__name__}({signature})")
        value = func(*args, **kwargs)
        print(f"{func.__name__!r} returned {value!r}")  # 4

        return value

    return wrapper_debug



from pathlib import Path
from typing import Any, List, Set, Union

def chdir(directory: Union[str, Path]) -> Any: ...
def remove_util(p: Union[str, Path]) -> None: ...

class PNGProvider:
    bitmaps_dir: Path
    def __init__(self, bitmaps_dir: Union[str, Path]) -> None: ...
    def get(self, key: str) -> Union[List[Path], Path]: ...

def add_missing_xcursors(directory: Path, data: List[Set[str]]=..., rename: bool=..., force: bool=...) -> None: ...
def timer(func: Any): ...
def debug(func: Any): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ctypes
import io
import math
import shlex
from ctypes import CDLL
from pathlib import Path
from struct import pack
from typing import Any, List, NamedTuple, Tuple

from PIL import Image, ImageFilter

from clickgen import __path__ as clickgen_pkg_root
from clickgen.util import remove_util

# Typings
Frame = Tuple[int, int, int, str, int]
Frames = List[Frame]
Color = Tuple[int, int, int, int]


class XCursor:
    """
    Build `XCursor` from the `.in` config file. This class is using `xcursorgen` internally.
    """

    config_file: Path
    prefix: Path
    out_dir: Path
    out: Path

    # main function ctypes define
    _lib_location: Path = Path(clickgen_pkg_root[0]) / "xcursorgen.so"
    _lib: CDLL = CDLL(str(_lib_location.absolute()))
    _LP_c_char = ctypes.POINTER(ctypes.c_char)
    _LP_LP_c_char = ctypes.POINTER(_LP_c_char)
    _lib.main.argtypes = (ctypes.c_int, _LP_LP_c_char)

    def __init__(self, config_file: Path, out_dir: Path) -> None:
        if not config_file.exists() or not config_file.is_file():
            raise FileNotFoundError(
                f"'{config_file.name}' is not found or not a config file"
            )

        self.config_file: Path = config_file
        self.prefix: Path = config_file.parent

        self.out_dir: Path = out_dir / "cursors"

        self.out_dir.mkdir(parents=True, exist_ok=True)

        self.out: Path = self.out_dir / self.config_file.stem

    def gen_argv_ctypes(self, argv: List[str]) -> Any:
        """ Convert `string` arguments to `ctypes` pointer. """
        p = (self._LP_c_char * len(argv))()

        for i, arg in enumerate(argv):
            enc_arg: bytes = str(arg).encode("utf-8")
            p[i] = ctypes.create_string_buffer(enc_arg)

        return ctypes.cast(p, self._LP_LP_c_char)

    def generate(self) -> None:
        """ Generate x11 cursor from `.in` file."""

        # remove old cursor file
        remove_util(self.out)

        argv: List[str] = [
            "xcursorgen",
            "-p",  # prefix args for xcursorgen (do not remove)
            str(self.prefix.absolute()),  # prefix args for xcursorgen (do not remove)
            str(self.config_file.absolute()),  # cursor's config/alias file
            str(self.out.absolute()),  # xcursor/output path
        ]

        kwargs: ctypes.pointer[ctypes.c_char] = self.gen_argv_ctypes(argv)
        args: ctypes.c_int = ctypes.c_int(len(argv))

        exec_with_error: bool = bool(self._lib.main(args, kwargs))
        if exec_with_error:
            raise RuntimeError(
                f"'xcursorgen' failed to generate XCursor from '{self.config_file.name}'"
            )

    @classmethod
    def create(cls, alias_file: Path, out_dir: Path) -> Path:
        cursor = cls(alias_file, out_dir)
        cursor.generate()
        return cursor.out


class AnicursorgenArgs(NamedTuple):
    """
    Structure `anicursorgen.py` CLI arguments.

    @add_shadow : Do not generate shadows for cursors (assign False to cancel its effect).

    @blur: Blur radius, in percentage of the canvas size (default is 3.125, set to 0 to disable blurring).

    @color: Shadow color in (RR,GG,BB,AA) (default is (0, 0, 0, 64)).

    @down_shift: Shift shadow down by this percentage of the canvas size (default is 3.125).

    @right_shift: Shift shadow right by this percentage of the canvas size (default is 9.375).
    """

    add_shadows: bool = False
    blur: float = 3.125
    color: Color = (0, 0, 0, 64)
    down_shift: float = 3.125
    right_shift: float = 9.375


class WindowsCursor:
    """
    Build Windows cursors from `.in` configs files. Code inspiration from `anicursorgen.py`.

    anicursorgen

    https://github.com/ubuntu/yaru/blob/master/icons/src/cursors/anicursorgen.py

    Copyright (C) 2015 Руслан Ижбулатов <lrn1986@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    """

    args: AnicursorgenArgs
    config_file: Path
    prefix: Path
    out_dir: Path
    out: Path

    def __init__(self, config_dir: Path, out_dir: Path, args: AnicursorgenArgs) -> None:
        self.config_file = config_dir
        self.prefix = config_dir.parent
        self.out_dir = out_dir
        self.args = args

        self.out_dir.mkdir(exist_ok=True, parents=True)

    def get_frames(self) -> Frames:
        in_buffer = self.config_file.open("rb")
        frames: Frames = []

        for line in in_buffer.readlines():
            line = line.decode()
            words = shlex.split(line.rstrip("\n").rstrip("\r"))

            size = int(words[0])
            xhot = int(words[1]) - 1
            yhot = int(words[2]) - 1
            filename = words[3]
            if not Path(filename).is_absolute():
                filename = str(self.prefix / filename)

            if len(words) > 4:
                duration = int(words[4])
            else:
                duration = 0

            frames.append((size, xhot, yhot, filename, duration))

        in_buffer.close()
        return frames

    @staticmethod
    def frames_have_animation(frames: Frames) -> bool:
        sizes = set()
        for frame in frames:
            if frame[4] == 0:
                continue
            if frame[0] in sizes:
                return True
            sizes.add(frame[0])

        return False

    @staticmethod
    def make_framesets(frames: Frames) -> Frames:
        framesets: Frames = []
        sizes = set()

        # This assumes that frames are sorted
        size: int = 0
        counter: int = 0
        for i, frame in enumerate(frames):

            if size == 0 or frame[0] != size:
                counter = 0
                size = frame[0]

                if size in sizes:
                    raise ValueError(
                        f"Frames are not sorted: frame {i} has size {size}, but we have seen that already"
                    )

                sizes.add(size)

            if counter >= len(framesets):
                framesets.append([])

            framesets[counter].append(frame)
            counter += 1

        for i in range(1, len(framesets)):
            if len(framesets[i - 1]) != len(framesets[i]):
                raise ValueError(
                    f"Frameset {i} has size {len(framesets[i])}, expected {len(framesets[i - 1])}"
                )

        for frameset in framesets:
            for i in range(1, len(frameset)):
                if frameset[i - 1][4] != frameset[i][4]:
                    raise ValueError(
                        f"Frameset {i} has duration {int(frameset[i][4])} for framesize {int(frameset[i][0])}, but {int(frameset[i - 1][4])} for framesize {int(frameset[i - 1][0])}",
                    )
        framesets = sorted(framesets, reverse=True)

        return framesets

    @staticmethod
    def copy_to(out: io.BufferedWriter, buf: io.BytesIO) -> None:
        buf.seek(0, io.SEEK_SET)
        while True:
            b = buf.read(1024)
            if len(b) == 0:
                break
            out.write(b)

    def make_ani(
        self,
        frames: Frames,
        out_buffer: io.BufferedWriter,
    ) -> None:
        framesets = self.make_framesets(frames)

        buf = io.BytesIO()

        buf.write(b"RIFF")
        riff_len_pos = buf.seek(0, io.SEEK_CUR)
        buf.write(pack("<I", 0))
        riff_len_start = buf.seek(0, io.SEEK_CUR)

        buf.write(b"ACON")
        buf.write(b"anih")
        buf.write(
            pack(
                "<IIIIIIIIII",
                36,
                36,
                len(framesets),
                len(framesets),
                0,
                0,
                32,
                1,
                int(framesets[0][0][4]),
                0x01,
            )
        )

        rates = set()
        for frameset in framesets:
            rates.add(int(frameset[0][4]))

        if len(rates) != 1:
            buf.write(b"rate")
            buf.write(pack("<I", len(framesets) * 4))
            for frameset in framesets:
                buf.write(pack("<I", int(frameset[0][4])))

        buf.write(b"LIST")
        list_len_pos = buf.seek(0, io.SEEK_CUR)
        buf.write(pack("<I", 0))
        list_len_start = buf.seek(0, io.SEEK_CUR)

        buf.write(b"fram")

        for frameset in framesets:
            buf.write(b"icon")
            cur = self.make_cur(frameset, animated=True)
            cur_size = cur.seek(0, io.SEEK_END)
            # aligned_cur_size = cur_size
            # if cur_size % 4 != 0:
            #  aligned_cur_size += 4 - cur_size % 2
            buf.write(pack("<i", cur_size))
            self.copy_to(buf, cur)
            pos = buf.seek(0, io.SEEK_END)
            if pos % 2 != 0:
                buf.write(("\x00" * (2 - (pos % 2))).encode())

        end_at = buf.seek(0, io.SEEK_CUR)
        buf.seek(riff_len_pos, io.SEEK_SET)
        buf.write(pack("<I", end_at - riff_len_start))
        buf.seek(list_len_pos, io.SEEK_SET)
        buf.write(pack("<I", end_at - list_len_start))

        self.copy_to(out_buffer, buf)

    @staticmethod
    def shadowize(shadow: Image, orig: Image, color: Color) -> None:
        o_pxs = orig.load()
        s_pxs = shadow.load()
        for y in range(orig.size[1]):
            for x in range(orig.size[0]):
                o_px = o_pxs[x, y]
                if o_px[3] > 0:
                    s_pxs[x, y] = (
                        int(color[0]),
                        int(color[1]),
                        int(color[2]),
                        int(color[3] * (o_px[3] / 255.0)),
                    )

    def create_shadow(self, orig: Image) -> Tuple[int, Any]:
        blur_px = orig.size[0] / 100.0 * self.args.blur
        right_px = int(orig.size[0] / 100.0 * self.args.right_shift)
        down_px = int(orig.size[1] / 100.0 * self.args.down_shift)

        shadow = Image.new("RGBA", orig.size, (0, 0, 0, 0))
        self.shadowize(shadow, orig, self.args.color)
        shadow.load()

        if self.args.blur > 0:
            crop = (
                int(math.floor(-blur_px)),
                int(math.floor(-blur_px)),
                orig.size[0] + int(math.ceil(blur_px)),
                orig.size[1] + int(math.ceil(blur_px)),
            )
            right_px += int(math.floor(-blur_px))
            down_px += int(math.floor(-blur_px))
            shadow = shadow.crop(crop)
            flt = ImageFilter.GaussianBlur(blur_px)
            shadow = shadow.filter(flt)
        shadow.load()

        shadowed: Image = Image.new("RGBA", orig.size, (0, 0, 0, 0))
        shadowed.paste(shadow, (right_px, down_px))
        shadowed.crop((0, 0, orig.size[0], orig.size[1]))
        shadowed = Image.alpha_composite(shadowed, orig)

        return (0, shadowed)

    @staticmethod
    def write_png(out: io.BufferedWriter, frame_png: Image) -> None:
        frame_png.save(out, "png", optimize=True)

    @staticmethod
    def write_cur(out: io.BufferedWriter, frame: Frame, frame_png: Image) -> None:
        pixels = frame_png.load()

        out.write(
            pack("<I II HH IIIIII", 40, frame[0], frame[0] * 2, 1, 32, 0, 0, 0, 0, 0, 0)
        )

        for y in reversed(list(range(frame[0]))):
            for x in range(frame[0]):
                pixel = pixels[x, y]
                out.write(pack("<BBBB", pixel[2], pixel[1], pixel[0], pixel[3]))

        acc = 0
        acc_pos = 0
        for y in reversed(list(range(frame[0]))):
            wrote = 0
            for x in range(frame[0]):
                if pixels[x, y][3] <= 127:
                    acc = acc | (1 << acc_pos)
                acc_pos += 1
                if acc_pos == 8:
                    acc_pos = 0
                    out.write(chr(acc).encode())
                    wrote += 1
            if wrote % 4 != 0:
                out.write(b"\x00" * (4 - wrote % 4))

    def make_cur(self, frames: Frames, animated: bool = False) -> io.BytesIO:
        buf = io.BytesIO()
        buf.write(pack("<HHH", 0, 2, len(frames)))
        frame_offsets = []

        frames = sorted(frames, reverse=True)

        for frame in frames:
            width = frame[0]
            if width > 255:
                width = 0
            height = width

            a = 0 if frame[1] == -1 else frame[1]
            b = 0 if frame[2] == -1 else frame[2]

            buf.write(pack("<BBBB HH", width, height, 0, 0, a, b))
            size_offset_pos = buf.seek(0, io.SEEK_CUR)

            buf.write(pack("<II", 0, 0))
            frame_offsets.append([size_offset_pos])

        for i, frame in enumerate(frames):
            frame_offset = buf.seek(0, io.SEEK_CUR)
            frame_offsets[i].append(frame_offset)

            frame_png = Image.open(frame[3])

            if self.args.add_shadows:
                succeeded, shadowed = self.create_shadow(frame_png)
                if succeeded == 0:
                    frame_png.close()
                    frame_png = shadowed

            #   Windows 10 fails to read PNG-compressed cursors for some reason
            #   and the information about storing PNG-compressed cursors is
            #   sparse. This is why PNG compression is not used.
            #   Previously this was conditional on cursor size (<= 48 to be uncompressed).
            compressed = False

            #   On the other hand, Windows 10 refuses to read very large
            #   uncompressed animated cursor files, but does accept
            #   PNG-compressed animated cursors for some reason. Go figure.
            if animated:
                compressed = True

            if compressed:
                self.write_png(buf, frame_png)
            else:
                self.write_cur(buf, frame, frame_png)

            frame_png.close()

            frame_end = buf.seek(0, io.SEEK_CUR)
            frame_offsets[i].append(frame_end - frame_offset)

        for frame_offset in frame_offsets:
            buf.seek(frame_offset[0])
            buf.write(pack("<II", frame_offset[2], frame_offset[1]))

        return buf

    def generate(self) -> None:
        frames = self.get_frames()
        animated = self.frames_have_animation(frames)

        name = self.config_file.stem
        ani_name = f"{name}.ani"
        cur_name = f"{name}.cur"

        cursor = ani_name if animated else cur_name
        self.out = self.out_dir / cursor

        # Remove Windows cursor, Which has the same 'name' inside 'output' directory.
        # This is useful for rebuilding the cursor to 'semi-animated'.
        # Remove only one extension can generate issues in the `WindowsPackager` module.
        remove_util(self.out_dir / ani_name)
        remove_util(self.out_dir / cur_name)

        if animated:
            with self.out.open("wb") as out:
                self.make_ani(frames, out)
        else:
            with self.out.open("wb") as out:
                buf = self.make_cur(frames)
                self.copy_to(out, buf)

    @classmethod
    def create(cls, alias_file: Path, out_dir: Path, args=AnicursorgenArgs()) -> Path:
        cursor = cls(alias_file, out_dir, args)
        cursor.generate()
        return cursor.out



import io
from PIL import Image
from pathlib import Path
from typing import Any, List, NamedTuple, Tuple

Frame = Tuple[int, int, int, str, int]
Frames = List[Frame]
Color = Tuple[int, int, int, int]

class XCursor:
    config_file: Path
    prefix: Path
    out_dir: Path
    out: Path
    def __init__(self, config_file: Path, out_dir: Path) -> None: ...
    def gen_argv_ctypes(self, argv: List[str]) -> Any: ...
    def generate(self) -> None: ...
    @classmethod
    def create(cls: Any, alias_file: Path, out_dir: Path) -> Path: ...

class AnicursorgenArgs(NamedTuple):
    add_shadows: bool = ...
    blur: float = ...
    color: Color = ...
    down_shift: float = ...
    right_shift: float = ...

class WindowsCursor:
    args: AnicursorgenArgs
    config_file: Path
    prefix: Path
    out_dir: Path
    out: Path
    def __init__(self, config_dir: Path, out_dir: Path, args: AnicursorgenArgs) -> None: ...
    def get_frames(self) -> Frames: ...
    @staticmethod
    def frames_have_animation(frames: Frames) -> bool: ...
    @staticmethod
    def make_framesets(frames: Frames) -> Frames: ...
    @staticmethod
    def copy_to(out: io.BufferedWriter, buf: io.BytesIO) -> None: ...
    def make_ani(self, frames: Frames, out_buffer: io.BufferedWriter) -> None: ...
    @staticmethod
    def shadowize(shadow: Image, orig: Image, color: Color) -> None: ...
    def create_shadow(self, orig: Image) -> Tuple[int, Any]: ...
    @staticmethod
    def write_png(out: io.BufferedWriter, frame_png: Image) -> None: ...
    @staticmethod
    def write_cur(out: io.BufferedWriter, frame: Frame, frame_png: Image) -> None: ...
    def make_cur(self, frames: Frames, animated: bool=...) -> io.BytesIO: ...
    def generate(self) -> None: ...
    @classmethod
    def create(cls: Any, alias_file: Path, out_dir: Path, args: Any=...) -> Path: ...



from pathlib import Path
from typing import Dict, List, Optional, Set

Data = List[Set[str]]
DATA: Data

class CursorDB:
    data: Dict[str, List[str]] = ...
    def __init__(self, data: Data) -> None: ...
    def search_symlinks(self, key: str, find_similar: bool=...) -> Optional[List[str]]: ...
    def rename_file(self, p: Path) -> Optional[Path]: ...



from PIL.Image import Image as Image
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

Size = Tuple[int, int]
LikePath = Union[str, Path]
LikePathList = Union[List[str], List[Path]]

class Bitmap:
    animated: bool
    png: Path
    grouped_png: List[Path]
    key: str
    x_hot: int
    y_hot: int
    size: Tuple[int, int]
    width: int
    height: int
    compress: int = ...
    def __init__(self, png: Union[LikePath, LikePathList], hotspot: Tuple[int, int]) -> None: ...
    def __enter__(self) -> Bitmap: ...
    def __exit__(self, exception_type: Any, exception_value: Any, traceback: Any) -> None: ...
    def resize(self, size: Size, resample: int=..., save: bool=...) -> Optional[Union[Image, List[Image]]]: ...
    def reproduce(self, size: Size=..., canvas_size: Size=..., position: str=..., save: Any=...) -> Optional[Union[Image, List[Image]]]: ...
    def rename(self, key: str) -> None: ...
    def copy(self, path: Optional[LikePath]=...) -> Bitmap: ...

class CursorAlias:
    bitmap: Bitmap
    prefix: str
    alias_dir: Path
    alias_file: Path
    garbage_dirs: List[Path] = ...
    def __init__(self, bitmap: Bitmap) -> None: ...
    def __enter__(self) -> CursorAlias: ...
    def __exit__(self, exception_type: Any, exception_value: Any, traceback: Any) -> None: ...
    @classmethod
    def from_bitmap(cls: Any, png: Union[LikePath, LikePathList], hotspot: Tuple[int, int]) -> CursorAlias: ...
    def create(self, sizes: Union[Size, List[Size]], delay: int=...) -> Path: ...
    def check_alias(self) -> None: ...
    def extension(self, ext: Optional[str]=...) -> LikePath: ...
    def copy(self, dst: Optional[LikePath]=...) -> CursorAlias: ...
    def rename(self, key: str) -> Path: ...
    def reproduce(self, size: Size=..., canvas_size: Size=..., position: str=..., delay: int=...) -> CursorAlias: ...



from pathlib import Path
from string import Template
from typing import Any, Dict, Optional, Set

THEME_FILES_TEMPLATES: Dict[str, Template]

def XPackager(directory: Path, theme_name: str, comment: str) -> None: ...

INSTALL_INF: Any
REQUIRED_WIN_CURSORS: Set[str]

def WindowsPackager(directory: Path, theme_name: str, comment: str, author: str, website_url: Optional[str]=...) -> None: ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import shutil
from copy import deepcopy
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Optional, Tuple, Union

from PIL import Image as Img
from PIL.Image import Image

from clickgen.util import remove_util

# Typing
Size = Tuple[int, int]
LikePath = Union[str, Path]
LikePathList = Union[List[str], List[Path]]


class Bitmap(object):
    """The ``Bitmap`` class is used to represent a ``.png`` or sequences of 
    ``.png`` image. The class also provides a number of factory functions, \
    including functions to **rename** ``.png`` images from files, and \
    **reproduce** size of same image/s.
    """

    animated: bool
    png: Path
    grouped_png: List[Path]

    key: str

    x_hot: int
    y_hot: int

    size: Tuple[int, int]
    width: int
    height: int

    compress: int = 0

    def __init__(
        self,
        png: Union[LikePath, LikePathList],
        hotspot: Tuple[int, int],
    ) -> None:
        """
        :param png: File location. Use ``List`` for animated Cursor.
        :type png: Union[LikePath, LikePathList]

        :param hotspot: Hotspot is coordinate value in Tuple. Cursor change \
                        state is calculated from this value.
        :type hotspot: Tuple[int, int]

        :raise TypeError: If provided ``.png`` file/s location is not type \
                          **str** or **pathlib.Path**
        """
        super().__init__()

        # Is png == _P             => 'static' bitmap
        # Or png == [_P, _P, ...]  => 'animated' bitmap
        # Or png == [_P]           => 'static' bitmap
        # else TypeError()

        err: str = f"argument should be a 'str' object or 'Path' object , not \
              {type(png)}"

        if isinstance(png, (str, Path)):
            self.__set_as_static(png, hotspot)

        elif isinstance(png, list):
            if len(png) == 1:
                self.__set_as_static(png[0], hotspot)
            else:
                self.__set_as_animated(png, hotspot)
        else:
            raise TypeError(err)

    def __str__(self) -> str:
        common: str = f"key={self.key}, animated={self.animated}, size={self.size}, width={self.width}, height={self.height}, x_hot={self.x_hot}, y_hot={self.y_hot}"
        if self.animated:
            return f"Bitmap(grouped_png={self.grouped_png}, {common})"
        return f"Bitmap(png={self.png}, {common})"

    def __repr__(self) -> str:
        common: str = f"'key':'{self.key}', 'animated':{self.animated}, 'size':{self.size}, 'width':{self.width}, 'height':{self.height}, 'x_hot':{self.x_hot}, 'y_hot':{self.y_hot}"
        if self.animated:
            return f"{{ 'grouped_png':{self.grouped_png}, {common} }}"
        return f"{{ 'png':{self.png}, {common} }}"

    # Context manager support
    def __enter__(self) -> "Bitmap":
        return self

    def __exit__(self, exception_type, exception_value, traceback):  # type: ignore
        return

    #
    # Private methods
    #
    def __set_as_static(self, png: LikePath, hotspot: Tuple[int, int]) -> None:
        """Set this Bitmap as **static**, It means this bitmap hold single ``png``.

                .. note:: This method called by ``self.__init__``.

        :param png: ``.png`` file location.
        :type png: Union[str,Path]

        :param hotspot: Hotspot is coordinate value in Tuple. Cursor change \
                        state is calculated from this value.
        :type hotspot: Tuple[int, int]

        :raise FileNotFoundError: If ``.png`` file not found
        :raise ValueError: If provided bitmap is not ``.png``
        :raise ValueError: If image width & height are not same
        :raise ValueError: If grouped ``.png`` files naming is invalid
        :raise ValueError: If hotspot is set higher or lower then image pixels
        """
        self.png = self._check_bitmap(png)
        self._set_key(self.png, check=False)
        self._set_size(self.png)
        self._set_hotspot(self.png, hotspot)
        self.animated = False

    def __set_as_animated(self, png: LikePathList, hotspot: Tuple[int, int]) -> None:
        """Set this Bitmap as **animated**, It means this bitmap holds multiple ``png``.

                .. note:: This method called by ``self.__init__``.

        :param png: ``.png`` file location ***List***.
        :type png: List[Union[str,Path]]

        :param hotspot: Hotspot is coordinate value in Tuple. Cursor change \
                        state is calculated from this value.
        :type hotspot: Tuple[int, int]

        :raise FileNotFoundError: If ``.png`` file not found
        :raise ValueError: If provided bitmap is not ``.png``
        :raise ValueError: If image width & height are not same
        :raise ValueError: If grouped ``.png`` files naming is invalid
        :raise ValueError: If hotspot is set higher or lower then image pixels
        """
        self.grouped_png = []
        for p in png:
            frame: Path = self._check_bitmap(p)

            self.grouped_png.append(frame)
            self._set_key(frame, check=True)
            self._set_size(frame)

        self.grouped_png.sort()

        # Don't worry, All images sizes are equal.
        self._set_hotspot(self.grouped_png[0], hotspot)
        self.animated = True

    #
    # Protected methods
    #
    def _check_bitmap(self, bmp_path: LikePath) -> Path:
        """Checks bitmap is supported by clickgen.

        :param bmp_path: Bitmap file location.
        :type bmp_path: Union[str, Path]

        :raise FileNotFoundError: If ``.png`` file not found
        :raise ValueError: If provided bitmap is not ``.png``

        :returns: Return the ``pathlib.Path`` instant, If provided **bitmap path** checks passed.
        :rtype: Path
        """
        p: Path = Path(bmp_path)
        if not p.exists():
            raise FileNotFoundError(
                f"Not a such file '{p.name}' in '{p.parent.absolute()}'"
            )

        # Supported bitmap type
        # => *.png
        for bmp_pattern in ("*.png",):
            if not p.match(bmp_pattern):
                raise ValueError(
                    f"{self.__class__} supports '{bmp_pattern}' bitmaps type, not '{p.suffix}'"
                )
        return p

    def _set_size(self, bmp_path: Path, prev_check: bool = True) -> None:
        """Set or overwrite size of this bitmap.

        :param bmp_path: Bitmap file location.
        :type bmp_path: Path

        :param prev_check: If you want to match size with previous size. This flag is useful for grouped ``png``.(@default True )
        :type prev_check: bool

        :raise ValueError: If image width & height are not same
        """
        with Img.open(bmp_path) as i:

            def __set() -> None:
                self.size = i.size
                self.width = i.width
                self.height = i.height

            if i.width == i.height:
                if prev_check:

                    try:
                        if self.size != i.size:
                            raise ValueError("All .png file's size must be equal")
                        __set()
                    except AttributeError:
                        __set()

                else:
                    __set()
            else:
                raise ValueError(
                    f"frame '{bmp_path.name}' must had equal width & height."
                )

    def _set_key(self, bmp_path: Path, check: bool) -> None:
        """Set unique identity for this bitmap.

        :param bmp_path: Bitmap file location.
        :type bmp_path: Path

        :param check: If you want sure multiple ``png`` files identity is same
        :type check: bool

        :raise ValueError: If grouped png indexing invalid.
        :raise ValueError: If new identity not matched with old.
        """
        if check:
            try:
                k, _ = bmp_path.stem.rsplit("-", 1)
            except ValueError:
                raise ValueError(
                    f"Invalid Bitmap name '{bmp_path.name}': Grouped Bitmaps must-have frame number followed by '-'. Like 'bitmap-000.png'"
                ) from None

            try:
                if self.key != k:
                    raise ValueError(
                        f"Bitmap '{bmp_path.name}' not matched with key '{self.key}'. Provide a Grouped Bitmaps with frame number followed by '-'.  Like 'bitmap-000.png','bitmap-001.png' "
                    )
                self.key = k
            except AttributeError:
                self.key = k
        else:
            self.key = bmp_path.stem

    def _set_hotspot(self, img_path: Path, hotspot: Tuple[int, int]) -> None:
        """Set this bitmap reaction state.

        :param img_path: Bitmap file location.
        :type img_path: Path

        :param hotspot: ``xy`` coordinates for this bitmap.
        :type hotspot: Tuple[int, int]
        """
        x = hotspot[0]
        y = hotspot[1]
        with Img.open(img_path) as i:
            if x > i.width or y > i.height:
                raise ValueError("'Hotspot' value is an overflow")

            if x < 0 or y < 0:
                raise ValueError("'Hotspot' value is an underflow")

            self.x_hot = x
            self.y_hot = y

    def _update_hotspots(self, new_size: Size) -> None:
        """Update this bitmap reaction state.

        :param new_size: Bitmap width & height tuple (in pixel)
        :type new_size: Tuple[int, int]
        """
        if self.size != new_size:
            self.x_hot = int(round(new_size[0] / self.width * self.x_hot))
            self.y_hot = int(round(new_size[1] / self.height * self.y_hot))

    #
    # Public methods
    #
    def resize(
        self,
        size: Size,
        resample: int = Img.NONE,
        save: bool = True,
    ) -> Optional[Union[Image, List[Image]]]:
        """Resize this bitmap.

        :param size: New width & height in pixel.
        :type size: Tuple[int, int]

        :param resample: Pillow resample algorithm.
        :type resample: int

        :param save: If you want to overwrite resized bitmap to actual png file. \
                     Neither it return pillow ``Image`` buffer.
        :type save: bool

        :returns: Returns image buffers, If *save* flag is set to ``False``
        :rtype: Optional[Union[Image, List[Image]]]

        :raise ValueError: If image width & height are not same.
        """

        def __resize(p: Path, frame: int) -> Image:
            img: Image = Img.open(p)

            # Preventing image quality degrades
            if img.size != size:
                img = img.resize(size, resample=resample)
                # If save => Update attribute
                if save:
                    self._set_size(p)
                    if frame == 0:
                        self._update_hotspots(size)
                    img.save(p, compress=self.compress)
            return img

        if self.animated:
            images: List[Image] = []
            for i, png in enumerate(self.grouped_png):
                img: Image = __resize(png, i)
                images.append(img)
            if not save:
                return images
            return None

        img: Image = __resize(self.png, 0)
        if not save:
            return img
        return None

    def reproduce(
        self,
        size: Size = (24, 24),
        canvas_size: Size = (32, 32),
        position: str = "center",
        save=True,
    ) -> Optional[Union[Image, List[Image]]]:
        """Resize bitmap with more options.

        :param size: Bitmap width & height in pixel.
        :type size: Tuple[int, int]

        :param canvas_size: Bitmap's canvas width & height in pixel.
        :type canvas_size: Tuple[int, int]

        :param position: Bitmap's canvas width & height in pixel. (@default "center")
        :type position: "center" | "top_left" | "top_right" | "bottom_left" | "bottom_right"

        :param save: If you want to overwrite resized bitmap to actual png \
                       file. Neither it return pillow ``Image`` buffer.
        :type save: bool

        :returns: Returns image buffers, If *save* flag is set to ``False``
        :rtype: Optional[Union[Image, List[Image]]]

        :raise ValueError: If image width & height are not same.
        """

        def __reproduce(p: Path) -> Image:
            frame: Image = Img.open(p).resize(size, resample=Img.BICUBIC)
            x, y = tuple(map(lambda i, j: i - j, canvas_size, size))

            switch: Dict[str, Tuple[int, int]] = {
                "top_left": (0, 0),
                "top_right": (x, 0),
                "bottom_left": (0, y),
                "bottom_right": (x, y),
                "center": (round(x / 2), round(y / 2)),
            }

            box: Tuple[int, int] = switch[position]

            canvas: Image = Img.new("RGBA", canvas_size, color=(256, 0, 0, 0))
            canvas.paste(frame, box=box)

            if save:
                canvas.save(p, compress=self.compress)
                self.x_hot = int(round(size[0] / self.width * self.x_hot) + box[0])
                self.y_hot = int(round(size[1] / self.height * self.y_hot) + box[1])
                self._set_size(p, prev_check=False)
            return canvas

        if self.animated:
            images: List[Image] = []
            for png in self.grouped_png:
                images.append(__reproduce(png))
            if not save:
                return images
            return None

        image: Image = __reproduce(self.png)
        if not save:
            return image
        return None

    def rename(self, key: str) -> None:
        """Rename unique identity of bitmap.

        :param key: Bitmap unique identity.
        :type key: str

        :raise ValueError: If grouped png indexing invalid.
        :raise ValueError: If new identity not matched with old.
        """
        old_key = self.key
        if key != old_key:

            def __rename(png: Path, check: bool) -> Path:
                name: str = f"{png.stem.replace(old_key, key, 1)}{png.suffix}"
                renamed_path: Path = png.with_name(name)
                png.rename(renamed_path)
                self._set_key(renamed_path, check)
                return renamed_path

            if self.animated:
                new_pngs = []
                self.key = key
                for png in self.grouped_png:
                    new_pngs.append(__rename(png, check=True))
                self.grouped_png = new_pngs
            else:
                self.png = __rename(self.png, check=False)

    def copy(self, path: Optional[LikePath] = None) -> "Bitmap":
        """Generate deepcopy of this bitmap.

        :param path: Provide custom path to store deepcopy bitmaps. \
                     Set ``None`` to store in temporary directory. \
                     (@default None)
        :type path: str

        :returns: deepcopy of this bitmap state.
        :rtype: Bitmap

        :raise NotADirectoryError: If provided path is not directory.
        """
        if not path:
            p_obj: Path = Path(mkdtemp(prefix=f"{self.key}__copy__"))
        else:
            p_obj: Path = Path(path)

        if p_obj.is_file():
            raise NotADirectoryError(f"path '{p_obj.absolute()}' is not a directory")

        p_obj.mkdir(parents=True, exist_ok=True)

        def __copy(src: Path) -> Path:
            dst: Path = p_obj / src.name
            shutil.copy2(src, dst)
            return dst

        if self.animated:
            pngs: List[Path] = []
            for p in self.grouped_png:
                pngs.append(__copy(p))
            return Bitmap(pngs, (self.x_hot, self.y_hot))

        p = __copy(self.png)
        return Bitmap(p, (self.x_hot, self.y_hot))


class CursorAlias(object):
    bitmap: Bitmap
    prefix: str
    alias_dir: Path
    alias_file: Path
    garbage_dirs: List[Path] = []

    """Cursor Config ``.in`` or ``.alias`` file provider."""

    def __init__(
        self,
        bitmap: Bitmap,
    ) -> None:
        """
        :param bitmap: Cursor :py:class:`~clickgen.core.Bitmap` instant.
        :type bitmap: Bitmap
        """
        super().__init__()

        self.bitmap = bitmap
        self.prefix = f"{self.bitmap.key}__alias"
        self.alias_dir = Path(mkdtemp(prefix=self.prefix))

    def __get_alias_file(self) -> Optional[Path]:
        """
        :returns: Return cursor alias file path, If it's exists.
        :rtype: Optional[Path]
        """
        if hasattr(self, "alias_file"):
            return self.alias_file
        return None

    def __str__(self) -> str:
        return f"CursorAlias(bitmap={self.bitmap!s}, prefix={self.prefix}, alias_dir={self.alias_dir}, alias_file={self.__get_alias_file()}, garbage_dirs={self.garbage_dirs})"

    def __repr__(self) -> str:
        return f"{{ 'bitmap':{self.bitmap!r}, 'prefix':{self.prefix}, 'alias_dir':{self.alias_dir}, 'alias_file':{self.__get_alias_file()}, 'garbage_dirs':{self.garbage_dirs} }}"

    # Context manager support
    def __enter__(self) -> "CursorAlias":
        return self

    def __exit__(self, exception_type, exception_value, traceback):  # type: ignore

        if hasattr(self, "alias_dir"):
            remove_util(self.alias_dir)

        if hasattr(self, "garbage_dirs"):
            for p in self.garbage_dirs:
                remove_util(p)

    @classmethod
    def from_bitmap(
        cls,
        png: Union[LikePath, LikePathList],
        hotspot: Tuple[int, int],
    ) -> "CursorAlias":
        """Create cursor alias config file from ``.png`` files instant.

        :param png: File location. Use ``List`` for animated Cursor.
        :type png: Union[LikePath, LikePathList]

        :param hotspot: Hotspot is coordinate value in Tuple. Cursor change \
                        state is calculated from this value.
        :type hotspot: Tuple[int, int]

        :raise TypeError: If provided ``.png`` file/s location is not type \
                          **str** or **pathlib.Path**
        """
        bmp: Bitmap = Bitmap(png, hotspot)
        return cls(bmp)

    def create(
        self,
        sizes: Union[Size, List[Size]],
        delay: int = 10,
    ) -> Path:
        """Generate and store cursor's config file at ``temporary`` storage.

        :param sizes: Cursor pixel size tuple.
        :type sizes: Union[Tuple[int, int], List[Tuple[int, int]]]

        :param delay: Delay between every cursor frame.(Affect on only \
                animated :py:class:`~clickgen.core.Bitmap`)
        :type delay: int

        :returns: Cursor alias file path.
        :rtype: Path

        :raise TypeError: If provided ``size`` is not type of Tuple[int, int] \
                or List of Tuple[int, int]
        """

        def __generate(size: Size) -> List[str]:
            d: Path = self.alias_dir / f"{size[0]}x{size[1]}"

            bmp: Bitmap = self.bitmap.copy(d)
            bmp.resize(size, resample=Img.BICUBIC)

            l: List[str] = []

            for file in d.glob("*.png"):
                fp: str = f"{file.relative_to(self.alias_dir)}"

                line: str = f"{size[0]} {bmp.x_hot} {bmp.y_hot} {fp}"
                if self.bitmap.animated:
                    line = f"{line} {delay}"

                l.append(f"{line}\n")

            return l

        def __write_alias(lines: List[str]) -> None:
            # sorting all lines according to size (24x24, 28x28, ..)
            lines.sort()
            # remove newline from EOF
            lines[-1] = lines[-1].rstrip("\n")

            cfg: Path = self.alias_dir / f"{self.bitmap.key}.alias"
            with cfg.open("w") as f:
                f.writelines(lines)
            self.alias_file = cfg

        sizes_type_err: str = (
            "argument 'sizes' should be Tuple[int, int] type or List[Tuple[int, int]]."
        )

        # Multiple sizes
        if isinstance(sizes, list):
            # Removing duplicate sizes
            sizes = sorted(set(sizes))

            lines: List[str] = []
            for size in sizes:
                if isinstance(size, tuple):
                    lines.extend(__generate(size))
                else:
                    raise TypeError(sizes_type_err)

            __write_alias(lines)

        # Single size
        elif isinstance(sizes, tuple):
            lines = __generate(sizes)
            __write_alias(lines)

        else:
            raise TypeError(sizes_type_err)

        return self.alias_file

    def check_alias(self) -> None:
        """Checks cursor  config file exists on filesystem.

        :raise FileNotFoundError:  If cursor config file not generate. \
                Calling :py:meth:`~clickgen.core.CursorAlias.create` to \
                recreate it.
        """
        if not any(self.alias_dir.iterdir()):
            raise FileNotFoundError("Alias directory is empty or not exists.")

    def extension(self, ext: Optional[str] = None) -> Union[str, Path]:
        """``get`` or ``change`` cursor config file extension. This method helps \
        to change default config extension ``.alias`` to ``.in``.

        :param ext: Provide custom cursor's config file extension. Like \
                    ``o.extension(ext = ".in")``. (@default None)
        :type ext: Optional[str]

        :returns: Provide ``None`` value to retrieve current extension. \
                Either returns the updated cursor's config file path.
        :rtype: Union[str, Path]
        """
        self.check_alias()
        if ext:
            new_path: Path = self.alias_file.with_suffix(ext)
            self.alias_file = self.alias_file.rename(new_path)
            return self.alias_file

        return self.alias_file.suffix

    def copy(self, dst: Optional[LikePath] = None) -> "CursorAlias":
        """Copy current bitmap config to directory.

             .. note::
                 This method **creates** the ``dst`` directory if not exists.

        :param dst: Custom directory path for store deepcopy of cursor config \
                file and it's bitmaps. Provide ``None`` value to store at \
                temporary storage.
        :type dst: Optional[Union[str, Path]]

        :returns: deepcopy object of current ``CursorAlias`` state.
        :rtype: CursorAlias

        :raise FileNotFoundError:  If cursor config file not generate. \
                Calling :py:meth:`~clickgen.core.CursorAlias.create` to \
                recreate it.
        :raise NotADirectoryError: If provided ``dst`` is not a directory.
        """
        self.check_alias()

        if not dst:
            dst = mkdtemp(prefix=self.prefix)
        dst = Path(dst)

        if dst.is_file():
            raise NotADirectoryError(f"path '{dst.absolute()}' is not a directory")

        replica_object = deepcopy(self)

        shutil.copytree(
            self.alias_dir, dst, dirs_exist_ok=True, copy_function=shutil.copy
        )
        replica_object.alias_dir = dst
        replica_object.prefix = dst.stem
        replica_object.alias_file = dst / self.alias_file.name

        return replica_object

    def rename(self, key: str) -> Path:
        """Rename current cursor's config file.

        Renaming cursor config is affects naming of bitmaps. \
        This method is calls :py:meth:`~clickgen.core.Bitmap.rename` \
        internaly for renaming bitmaps name.

        :param key: Unique identity.
        :type key: str

        :returns: Renamed config file path.
        :rtype: Path

        :raise FileNotFoundError:  If cursor config file not generate. \
                Calling :py:meth:`~clickgen.core.CursorAlias.create` to \
                recreate it.
        :raise ValueError: If grouped png indexing invalid.
        :raise ValueError: If new identity not matched with old.
        """
        self.check_alias()
        old_key: str = self.bitmap.key

        if old_key != key:
            # Setting new_prefix & renaming alias directory
            new_prefix = f"{key}__alias"
            new_alias_dir = self.alias_dir.with_name(
                self.alias_dir.name.replace(self.prefix, new_prefix)
            )
            shutil.move(f"{self.alias_dir.absolute()}", new_alias_dir)
            self.prefix = new_prefix
            self.alias_dir = new_alias_dir

            # Renaming content
            def __rename(p: Path) -> Path:
                name: str = f"{p.stem.replace(old_key, key, 1)}{p.suffix}"
                path: Path = p.with_name(name)
                return p.rename(path)

            for f in self.alias_dir.iterdir():
                if f.is_dir():
                    for png in f.glob("*.png"):
                        png = __rename(png)
                elif f.is_file or f.absolute() == self.alias_file.absolute():
                    updated_data: str = f.read_text().replace(old_key, key)
                    f.write_text(updated_data)
                    self.alias_file = __rename(f)

        return self.alias_file

    def reproduce(
        self,
        size: Size = (24, 24),
        canvas_size: Size = (32, 32),
        position: str = "center",
        delay: int = 3,
    ) -> "CursorAlias":
        """Resize cursor config and bitmap.

        :param size: Bitmap width & height in pixel.
        :type size: Tuple[int, int]

        :param canvas_size: Bitmap's canvas width & height in pixel.
        :type canvas_size: Tuple[int, int]

        :param position: Bitmap's canvas width & height in pixel. (@default "center")
        :type position: "center" | "top_left" | "top_right" | "bottom_left" | "bottom_right"

        :param save: If you want to overwrite resized bitmap to actual png \
                       file. Neither it return pillow ``Image`` buffer.
        :type save: bool

        :returns: Returns reproduced ``CursorAlias`` object.
        :rtype: CursorAlias

        :raise FileNotFoundError:  If cursor config file not generate. \
                Calling :py:meth:`~clickgen.core.CursorAlias.create` to \
                recreate it.
        :raise ValueError: If image width & height are not same.
        """
        self.check_alias()

        tmp_bitmaps_dir: Path = Path(mkdtemp(prefix=f"{self.prefix}__garbage_bmps__"))
        tmp_bitmap = self.bitmap.copy(tmp_bitmaps_dir)
        tmp_bitmap.reproduce(size, canvas_size, position)

        # Will Deleted, When __exit__ being called
        self.garbage_dirs.append(tmp_bitmaps_dir)

        cls: CursorAlias = CursorAlias(tmp_bitmap)
        cls.create(canvas_size, delay)

        return cls



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from string import Template
from typing import Dict, List, Optional, Set


# --- X11

THEME_FILES_TEMPLATES: Dict[str, Template] = {
    "cursor.theme": Template('[Icon Theme]\nName=$theme_name\nInherits="$theme_name"'),
    "index.theme": Template(
        '[Icon Theme]\nName=$theme_name\nComment=$comment\nInherits="hicolor"'
    ),
}


def XPackager(directory: Path, theme_name: str, comment: str) -> None:
    """This packager generates ``cursor.theme`` & ``index.theme`` files at ``directory``.

    :param directory: Path where ``.theme`` files save.
    :param directory: Path

    :param theme_name: Cursor theme name.
    :param theme_name: str

    :param comment: Extra information about this cursor theme.
    :param comment: str

    :returns: Nothing
    :rtype: None
    """

    # Writing all .theme files
    files: Dict[str, str] = {}
    for file, template in THEME_FILES_TEMPLATES.items():
        files[file] = template.safe_substitute(theme_name=theme_name, comment=comment)

    for f, data in files.items():
        fp: Path = directory / f
        fp.write_text(data)


# --- Windows

INSTALL_INF = Template(
    """[Version]
signature="$CHICAGO$"
$comment

[DefaultInstall]
CopyFiles = Scheme.Cur, Scheme.Txt
AddReg    = Scheme.Reg

[DestinationDirs]
Scheme.Cur = 10,"%CUR_DIR%"
Scheme.Txt = 10,"%CUR_DIR%"

[Scheme.Reg]
HKCU,"Control Panel\\Cursors\\Schemes","%SCHEME_NAME%",,"%10%\\%CUR_DIR%\\%pointer%,%10%\\%CUR_DIR%\\%help%,%10%\\%CUR_DIR%\\%work%,%10%\\%CUR_DIR%\\%busy%,%10%\\%CUR_DIR%\\%Cross%,%10%\\%CUR_DIR%\\%Text%,%10%\\%CUR_DIR%\\%Hand%,%10%\\%CUR_DIR%\\%Unavailiable%,%10%\\%CUR_DIR%\\%Vert%,%10%\\%CUR_DIR%\\%Horz%,%10%\\%CUR_DIR%\\%Dgn1%,%10%\\%CUR_DIR%\\%Dgn2%,%10%\\%CUR_DIR%\\%move%,%10%\\%CUR_DIR%\\%alternate%,%10%\\%CUR_DIR%\\%link%"

; -- Installed files

[Scheme.Cur]
"$Work"
"$Busy"
"$Default"
"$Help"
"$Link"
"$Move"
"$Diagonal_2"
"$Vertical"
"$Horizontal"
"$Diagonal_1"
"$Handwriting"
"$Cross"
"$IBeam"
"$Unavailiable"
"$Alternate"

[Strings]
CUR_DIR           = "Cursors\\$theme_name Cursors"
SCHEME_NAME       = "$theme_name Cursors"
pointer           = "$Default"
help              = "$Help"
work              = "$Work"
busy              = "$Busy"
cross             = "$Cross"
text              = "$IBeam"
hand              = "$Handwriting"
unavailiable      = "$Unavailiable"
vert              = "$Vertical"
horz              = "$Horizontal"
dgn1              = "$Diagonal_1"
dgn2              = "$Diagonal_2"
move              = "$Move"
alternate         = "$Alternate"
link              = "$Link"
"""
)

REQUIRED_WIN_CURSORS: Set[str] = {
    "Work",
    "Busy",
    "Default",
    "Help",
    "Link",
    "Move",
    "Diagonal_2",
    "Vertical",
    "Horizontal",
    "Diagonal_1",
    "Handwriting",
    "Cross",
    "IBeam",
    "Unavailiable",
    "Alternate",
}


def WindowsPackager(
    directory: Path,
    theme_name: str,
    comment: str,
    author: str,
    website_url: Optional[str] = None,
) -> None:
    """This packager generates ``install.inf`` files at ``directory``. Also, Cursor extensions is identified by its type (.cur/.ani).

    :param directory: Path where ``.theme`` files save.
    :param directory: Path

    :param theme_name: Cursor theme name.
    :param theme_name: str

    :param comment: Extra information about this cursor theme.
    :param comment: str

    :param author: Author name.
    :param author: str

    :param website_url: Website web address.(Useful for **bug reports**)
    :param author: str (, optional)

    :returns: Nothing
    :rtype: None
    """

    files: List[Path] = []

    for extensions in ("*.ani", "*.cur"):
        for i in sorted(directory.glob(extensions)):
            if i.stem in REQUIRED_WIN_CURSORS:
                files.append(i)

    cursors: Set[Path] = set(files)

    # Checking cursor files
    if not cursors:
        raise FileNotFoundError(
            f"Windows cursors '*.cur' or '*.ani' not found in '{directory}'"
        )

    if len(cursors) < len(REQUIRED_WIN_CURSORS):
        # Some cursors are missing
        c = set(map(lambda x: x.stem, cursors))
        missing = sorted(REQUIRED_WIN_CURSORS - set(c))
        raise FileNotFoundError(f"Windows cursors are missing {missing}")

    if website_url:
        comment = f"{comment}\n{website_url}"

    # Real magic of python
    # replace $Default => Default.ani | Default.cur (as file was provided)
    cursor_data: Dict[str, str] = {}
    for cur in cursors:
        cursor_data[cur.stem] = cur.name

    data: str = INSTALL_INF.safe_substitute(
        theme_name=theme_name, comment=comment, author=author, **cursor_data
    )

    # Store install.inf file
    install_inf: Path = directory / "install.inf"
    install_inf.write_text(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
from distutils.command.install import install as _install
from shutil import which

from setuptools import setup


def make_path() -> str:
    path = which("make")
    if not path:
        raise Exception("'make' command not found")
    return path


class install(_install):
    def run(self):
        subprocess.call([make_path(), "clean", "-C", "xcursorgen"])
        subprocess.call([make_path(), "-C", "xcursorgen"])
        _install.run(self)


# readme.md as long description
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="clickgen",
    version="1.1.9",
    author="Kaiz Khatri",
    author_email="kaizmandhu@gmail.com",
    description="The hassle-free cursor building toolbox 🧰",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ful1e5/clickgen",
    cmdclass={"install": install},
    scripts=["scripts/clickgen"],
    project_urls={
        "Source": "https://github.com/ful1e5/clickgen",
        "Funding": "https://www.paypal.me/kaizkhatri",
        "Changelog": "https://github.com/ful1e5/clickgen/blob/main/CHANGELOG.md",
    },
    install_requires=["Pillow>=8.1.1"],
    packages=["clickgen"],
    package_dir={"clickgen": "clickgen"},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Multimedia :: Graphics",
        "Topic :: System :: Operating System",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    python_requires=">=3.8",
    include_package_data=True,
    keywords=["cursor", "xcursor", "windows", "linux", "anicursorgen", "xcursorgen"],
    zip_safe=True,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Dict, List, Optional, Set
from difflib import SequenceMatcher as SM

# Typing
Data = List[Set[str]]

DATA: Data = [
    {
        "X_cursor",
        "x-cursor",
        "kill",
        "pirate",
    },
    {"kill", "pirate"},
    {"all-scroll", "fleur", "size_all"},
    {
        "bd_double_arrow",
        "c7088f0f3e6c8088236ef8e1e3e70000",
        "38c5dff7c7b8962045400281044508d2",
        "size_fdiag",
        "nwse-resize",
    },
    {"bottom_left_corner", "sw-resize"},
    {"bottom_right_corner", "se-resize"},
    {"bottom_side", "s-resize"},
    {"bottom_tee"},
    {"center_ptr"},
    {
        "circle",
        "forbidden",
    },
    {"context-menu"},
    {
        "1081e37283d90000800003c07f3ef6bf",
        "6407b0e94181790501fd1e167b474872",
        "b66166c04f8c3109214a4fbd64a50fc8",
        "08ffe1cb5fe6fc01f906f1c063814ccf",
        "copy",
    },
    {
        "cross",
        "cross_reverse",
        "diamond_cross",
    },
    {"crossed_circle", "03b6e0fcb3499374a867c041f52298f0", "not-allowed"},
    {"crosshair"},
    {"dnd-ask"},
    {"dnd-copy"},
    {"dnd-link", "alias"},
    {"dnd-move"},
    {
        "dnd-none",
        "closedhand",
        "208530c400c041818281048008011002",
        "fcf21c00b30f7e3f83fe0dfd12e71cff",
    },
    {"dnd_no_drop", "no-drop"},
    {"dotbox", "dot_box_mask", "draped_box", "icon", "target"},
    {
        "fcf1c3c7cd4491d801f1e1c78f100000",
        "50585d75b494802d0151028115016902",
        "fd_double_arrow",
        "nesw-resize",
        "size_bdiag",
    },
    {"grabbing"},
    {"hand"},
    {
        "hand1",
        "grab",
        "5aca4d189052212118709018842178c0",
        "openhand",
    },
    {
        "9d800788f1b08800ae810202380a0822",
        "e29285e634086352946a0e7090d73106",
        "hand2",
        "pointer",
        "pointing_hand",
    },
    {
        "left_ptr",
        "arrow",
        "default",
    },
    {
        "00000000000000020006000e7e9ffc3f",
        "08e8e1c95fe2fc01f976f1e063a24ccd",
        "3ecb610c1bf2410f44200f48c40d3599",
        "9116a3ea924ed2162ecab71ba103b17f",
        "left_ptr_watch",
        "progress",
    },
    {"left_side", "w-resize"},
    {"left_tee"},
    {
        "3085a0e285430894940527032f8b26df",
        "640fb0e74195791501fd1ed57b41487f",
        "a2a266d0498c3104214a47bd64ab0fc8",
        "0876e1c15ff2fc01f906f1c363074c0f",
        "link",
    },
    {"ll_angle"},
    {"lr_angle"},
    {
        "4498f0e0c1937ffe01fd06f973665830",
        "9081237383d90e509aa00f00170e968f",
        "move",
    },
    {
        "pencil",
        "draft",
    },
    {"plus", "cell"},
    {"pointer-move"},
    {
        "5c6cd98b3f3ebcb1f9c7f1c204630408",
        "d9ce0ab605698f320427677b458ad60b",
        "gumby",
        "help",
        "left_ptr_help",
        "question_arrow",
        "whats_this",
    },
    {
        "right_ptr",
        "draft_large",
        "draft_small",
    },
    {"right_side", "e-resize"},
    {"right_tee"},
    {"sb_down_arrow", "down-arrow"},
    {
        "028006030e0e7ebffc7f7070c0600140",
        "14fef782d02440884392942c11205230",
        "043a9f68147c53184671403ffa811cc5",
        "col-resize",
        "ew-resize",
        "h_double_arrow",
        "sb_h_double_arrow",
        "size-hor",
        "size_hor",
        "split_h",
    },
    {"sb_left_arrow", "left-arrow"},
    {"sb_right_arrow", "right-arrow"},
    {"sb_up_arrow", "up-arrow"},
    {
        "00008160000006810000408080010102",
        "c07385c7190e701020ff7ffffd08103c",
        "2870a09082c103050810ffdffffe0204",
        "double_arrow",
        "ns-resize",
        "row-resize",
        "sb_v_double_arrow",
        "size-ver",
        "size_ver",
        "split_v",
        "v_double_arrow",
    },
    {"tcross", "color-picker"},
    {"top_left_corner", "nw-resize"},
    {"top_right_corner", "ne-resize"},
    {"top_side", "n-resize"},
    {"top_tee"},
    {"ul_angle"},
    {"ur_angle"},
    {"vertical-text", "048008013003cff3c00c801001200000"},
    {
        "watch",
        "clock",
        "wait",
        "0426c94ea35c87780ff01dc239897213",
    },
    {"wayland-cursor"},
    {
        "xterm",
        "text",
        "ibeam",
    },
    {"zoom-in", "f41c0e382c94c0958e07017e42b00462"},
    {"zoom-out", "f41c0e382c97c0938e07017e42800402"},
]


class CursorDB(object):
    """The CursorDB class provides ``Xcursor`` data available on X11/Wayland
    /XWayland.
    """

    data: Dict[str, List[str]] = {}

    def __init__(self, data: Data) -> None:
        """Initiate CursorDB instance.

        This call ``self.__seed`` for setting dataset into ``Dict`` structure.

        :param data: Provide custom dataset of cursors. default to ``clickgen.db.DATA``
        :type data: List[Set[str]]

        :returns: Instance of ``CursorDB``

        :raises TypeError: If provided ``data`` is not type of ``List``
        :raises TypeError: If provided ``data``'s list member is not type of ``Set``
        """
        super().__init__()
        self.__seed(data)

    def __str__(self) -> str:
        string: List[str] = []
        for index, data in self.data.items():
            string.append(f"{index}={data}")
        return f"CursorDB({', '.join(string)})"

    def __repr__(self) -> str:
        return f"{self.data!r}"

    def __seed(self, data: Data) -> None:
        if not isinstance(data, list):
            raise TypeError("'data' is not type of 'List'")
        for item in sorted(data):
            if not isinstance(item, set):
                raise TypeError(f"Item '{item}' is not type of 'Set'")
            for i in sorted(item):
                symlinks: List[str] = list(item)
                symlinks.remove(i)
                self.data[i] = symlinks

    def search_symlinks(
        self, key: str, find_similar: bool = False
    ) -> Optional[List[str]]:
        """Retrieve similar implementation Xcursor from database.

        :param key: Name of XCursor
        :type key: str

        :param find_similar: If this flag set to ``True``, It return similar named ``XCursor`` from database.
        :type find_similar: bool

        :return: List of missing ``XCursor`` in string list or None.
        :rtype: Optional[List[str]]

        """
        if find_similar:
            key = self.__find_similar(key)

        return self.data.get(key)

    def __find_similar(self, key: str) -> str:
        compare_ratio: float = 0.5
        result: str = key
        for e in self.data.keys():
            ratio: float = SM(None, key.lower(), e.lower()).ratio()
            if ratio > compare_ratio:
                compare_ratio = ratio
                result = e

        return result

    def rename_file(self, p: Path) -> Optional[Path]:
        """Search similar named Xcursor in database. If name of XCursor is not
        correct this return ``pathlib.Path`` Object.

        :param p: Path to XCursor
        :type key: Path

        :return: Return renamed ``Path`` If XCursor name not correct.
        :rtype: Optional[Path]

        :raise FileNotFoundError: If file is not found in file system or it's a directory.
        """
        if not p.is_file():
            raise FileNotFoundError(f"'{p.absolute()}' is not file")
        key: str = self.__find_similar(p.stem)
        if key != p.stem:
            return p.with_name(f"{key}{p.suffix}")
        return None
